#
# Created by Yi MIN<yi@zen.ly>
# Copyright © 2017 Zenly. All rights reserved.
#
class WDA
  module FindElement
    ### Searching for elements

    # Find element with given type and value, return first found element
    # @param type [String, Symbol], value [Stirng]
    # @return [Hash]
    def find(type, value)
      wait = Wait.new(:timeout => @timeout) # seconds

      begin
        wait.until {
          @element = post('/element', { using: stringlize(type), value: value }) 
          @element['status'] == 0
        }
      ensure
        raise Error::NoSuchElementError, "Can't find '#{value}' with type '#{type}'" if @element['status'] != 0 
      end

      Element.new self, @element['value']
    end

    # Find all elements with given type and value, return in an Array
    # @param type [String, Symbol], value [Stirng]
    # @return [Array]
    def finds(type, value)
      wait = Wait.new(:timeout => @timeout) # seconds
      
      begin
        wait.until { 
          @elements = post('/elements', { using: stringlize(type), value: value })['value']
          @elements.length != 0
        }
      ensure
        raise Error::NoSuchElementError, "Can't find '#{value}' with type '#{type}'" if @elements.length == 0 
      end

      @elements.map { |element| Element.new self, element }
    end

    # Find child element by given type and value, return first found element
    # @param id [String] parent element id, type [String, Symbol], value [Stirng]
    # @return [Object] found element
    def find_subl_element(id, type, value)
      wait = Wait.new(:timeout => @timeout) # seconds
      begin
        wait.until {
          @element = post('/element/' + id + '/element', { using: stringlize(type), value: value })
          @element['status'] == 0
        }
      ensure
        raise Error::NoSuchElementError, "Can't find '#{value}' with type '#{type}'" if @element['status'] != 0 
      end

      Element.new self, @element['value']
    end

    # Find children elements by given type and value, return elements array
    # @param id [String] parent element id, type [String, Symbol], value [Stirng]
    # @return [Array<Element>]
    def find_subl_elements(id, type, value)
      wait = Wait.new(:timeout => @timeout) # seconds
      
      begin
        wait.until { 
          @result = post('/element/' + id + '/elements', { using: stringlize(type), value: value })
          @elements = @result['value']
          @result['status'] != 7
        }
      ensure
        raise Error::NoSuchElementError, "Can't find '#{value}' with type '#{type}'" if @result['status'] != 0 
      end

      @elements.map { |element| Element.new self, element }
    end

    # Search with given name
    # @param value [String]
    # @return [Hash]
    def name(value)
      find :name, value
    end

    # Search all element with given name
    # @param value [String]
    # @return [Array]
    def names(value)
      finds :name, value
    end

    # Search with given id
    # @param value [String]
    # @return [Hash]
    def id(value)
      find :id, value
    end

    # Search all element with given id
    # @param value [String]
    # @return [Array]
    def ids(value)
      finds :id, value
    end

    # Search with given accessibility_id
    # @param value [String]
    # @return [Hash]
    def accessibility_id(value)
      find :accessibility_id, value
    end

    # Search all element with given accessibility_id
    # @param value [String]
    # @return [Array]
    def accessibility_ids(value)
      finds :accessibility_id, value
    end

    # Search with given value (Complete value)
    # @param value [String]
    # @return [Hash]
    def text(value)
      find :link_text, "label=#{value}"
    end

    # Search with given value (Complete value)
    # @param value [String]
    # @return [Array]
    def texts(value)
      finds :link_text, "label=#{value}"
    end

    # Search with given partial value (partial link text)
    # @param value [String]
    # @return [Hash]
    def partial_text(value)
      find :partial_link_text, "label=#{value}"
    end

    # Search with given partial value (partial link text)
    # @param value [String]
    # @return [Array]
    def partial_texts(value)
      finds :partial_link_text, "label=#{value}"
    end

    # Search with class name
    # @param value [String] XCUIElementType*, class name in XCUITest, ex: class_names('Button') to search XCUIElementTypeButton 
    # @return [Hash] found XCUIElementType* elements
    def class_name(value)
      find :class_name, match(value)
    end
    
    # Search with class name
    # @param value [String] XCUIElementType*, class name in XCUITest, ex: class_names('Button') to search XCUIElementTypeButton 
    # @return [Array] contains all XCUIElementType* elements
    def class_names(value)
      finds :class_name, match(value)
    end

    # Search with xpath
    # @param value [String] 
    # example: 
    # "//XCUIElementTypeButton[@name='Share']", 
    # "//XCUIElementTypeTextField[@value='Value']"
    # @return [Array] contains all elements which match given xpath
    def xpath(value)
      finds :xpath, value
    end

    # Xpath search with element type, element attribute, element attribute value
    # @param type [String], attribute [String], value [String] 
    # example: 
    # xpath('Button', 'name', 'Share') for "//XCUIElementTypeButton[@name='Share']"
    # @return [Array] contains all elements which match given variables
    def xpath_search(type, attribute, value)
      finds :xpath, "//#{match(type)}[@#{attribute}=\"#{value}\"]"
    end

    # Search predicate string, return first found element
    # @param predicate_value [String] 'isWDVisible=1'
    # example: 'isWDAccessible=1', 'isWDEnabled=0'
    # @return [Hash]
    def predicate_text(predicate_value)
      find :predicate_string, predicate_value 
    end

    # Search with predicate string
    # @param predicate_value [String] 'isWDVisible=1'
    # example: 'isWDAccessible=1', 'isWDEnabled=0', 'isWDAccessibilityContainer=1'
    # @return [Array]
    def predicate_texts(predicate_value)
      finds :predicate_string, predicate_value 
    end

    # @return element's visible cells [Array]
    def visible_cell(id)
      get '/wda/element/' + id + '/getVisibleCells'
    end

    # Find button by given index or value
    # @param value [String, Integer] 
    # If value is integer then return the button at that index
    # If value is string then return the first button which contains given value
    # @return [Hash, Element]
    def button(value = 0)
      if value.is_a? Numeric
        finds(:xpath, "//XCUIElementTypeButton")[value]
      else
        find(:xpath, "//XCUIElementTypeButton[@name=\"#{value}\"]")
      end
    end

    # Find buttons by given index or value
    # @param value [String, Integer] 
    # If value is integer then return the all buttons
    # If value is string then return the all buttons which contain given value
    # @return [Array] All found buttons
    def buttons(value = nil)
      if value.nil?
        finds(:xpath, "//XCUIElementTypeButton")
      else
        finds(:xpath, "//XCUIElementTypeButton[@name=\"#{value}\"]")
      end
    end

    # Find the first Button
    # @return [Hash] button
    def first_button
      finds(:xpath, "//XCUIElementTypeButton").first
    end

    # Find the last Button
    # @return [Hash] button
    def last_button
      finds(:xpath, "//XCUIElementTypeButton").last
    end

    # Find textfield by given index or value
    # @param value [String, Integer]
    # If value is integer then return the textField at that index
    # If value is string then return the first textField which contains given value
    # @return [TextField]
    def textfield(value = 0)
      if value.is_a? Numeric
        finds(:xpath, "//XCUIElementTypeTextField")[value]
      else
        find(:xpath, "//XCUIElementTypeTextField[@value=\"#{value}\"]")
      end
    end

    # Find the all TextField.
    # @param value [String]
    # If value is integer then return all textFields
    # If value is string then return all textFields which contain given value
    # @return [Array]
    def textfields(value = nil)
      if value.nil?
        finds(:xpath, "//XCUIElementTypeTextField")
      else
        finds(:xpath, "//XCUIElementTypeTextField[@value=\"#{value}\"]")
      end
    end

    # Find the first TextField.
    # @return [TextField]
    def first_textfield
      finds(:xpath, "//XCUIElementTypeTextField").first
    end

    # Find the last TextField.
    # @return [TextField]
    def last_textfield
      finds(:xpath, "//XCUIElementTypeTextField").last
    end

    # Find secure_textfield by given index or value
    # @param value [String, Integer] 
    # If value is integer then return the secure_textField at that index
    # If value is string then return the first secure_textField which contains given value
    # @return [TextField]
    def secure_textfield(value = 0)
      if value.is_a? Numeric
        finds(:xpath, "//XCUIElementTypeSecureTextField")[value]
      else
        find(:xpath, "//XCUIElementTypeSecureTextField[@value=\"#{value}\"]")
      end
    end

    # Find the all SecureTextField.
    # @return [Array]
    def secure_textfields 
      finds(:xpath, "//XCUIElementTypeSecureTextField")
    end

    # Find StaticText by given index or value
    # @param value [String, Integer]
    # If value is integer then return the StaticText at that index
    # If value is string then return the first StaticText which contains given value
    # @return [TextField]
    def statictext(value = 0)
      if value.is_a? Numeric
        finds(:xpath, "//XCUIElementTypeStaticText")[value]
      else
        find(:xpath, "//XCUIElementTypeStaticText[@value=\"#{value}\"]")
      end
    end

    # Find the all StaticText.
    # @param value [String]
    # If value is integer then return all StaticTexts
    # If value is string then return the all StaticTexts which contain given value
    # @return [Array]
    def statictexts(value = nil)
      if value.nil?
        finds(:xpath, "//XCUIElementTypeStaticText")
      else
        finds(:xpath, "//XCUIElementTypeStaticText[@value=\"#{value}\"]")
      end
    end

    # Find icon by given index or value
    # @param value [String, Integer] 
    # If value is integer then return the icon at that index
    # If value is string then return the first icon which contains given value
    # @return [Hash, Element]
    def icon(value = 0)
      if value.is_a? Numeric
        finds(:xpath, "//XCUIElementTypeIcon")[value]
      else
        find(:xpath, "//XCUIElementTypeIcon[@name=\"#{value}\"]")
      end
    end

    # Find icons by given index or value
    # @param value [String] 
    # If value is integer then return all icons
    # If value is string then return the all icons which contain given value
    # @return [Array] All found icons
    def icons(value = nil)
      if value.nil?
        finds(:xpath, "//XCUIElementTypeIcon")
      else
        finds(:xpath, "//XCUIElementTypeIcon[@name=\"#{value}\"]")
      end
    end

    # Find SearchField by given index or value
    # @param value [String, Integer] 
    # If value is integer then return the SearchField at that index
    # If value is string then return the first SearchField which has given value as name
    # @return [Hash, Element]
    def searchfield(value = 0)
      if value.is_a? Numeric
        finds(:xpath, "//XCUIElementTypeSearchField")[value]
      else
        find(:xpath, "//XCUIElementTypeSearchField[@name=\"#{value}\"]")
      end
    end

    # Find icons by given index or value
    # @param value [String] 
    # If value is integer then return all SearchFields
    # If value is string then return the all SearchFields which have given value as name
    # @return [Array] All found SearchFields
    def searchfields(value = nil)
      if value.nil?
        finds(:xpath, "//XCUIElementTypeSearchField")
      else
        finds(:xpath, "//XCUIElementTypeSearchField[@name=\"#{value}\"]")
      end
    end

    ########### This is calling selenium find_element* methods##########
    # Find element with given type and value, return first found element
    # @param args [*args]
    # @return [Object]
    def find_element(*args)
      @driver.find_element(*args)
    end

    # Find elements with given type and value, return in an Array
    # @param args [*args]
    # @return [Array<Element>]
    def find_elements(*args)
      @driver.find_elements(*args)
    end
    ####################################################################

    # Turn symbol to string, replace '_' to ' '
    # @param instring [Symbol]
    # @return [String]
    def stringlize(instring)
      if instring.is_a? Symbol
        instring.to_s.gsub('_', ' ')
      elsif instring.is_a? String
        return instring.gsub('_', ' ')
      else 
        fail 'Xpath searching type should be a String'
      end
    end 

    def set_timeout(second)
      @timeout = second
    end
  end
end